"""
Module initialization_handler.py from the package PyZEAL.
This module handles initialization for the default PyZEAL services.

Authors:\n
- Philipp Schuette\n
"""

from typing import Optional

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.cli.cli_controller import CLIController
from pyzeal.cli.cli_parser import PyZEALParser
from pyzeal.cli.controller_facade import CLIControllerFacade
from pyzeal.cli.parser_facade import PyZEALParserInterface
from pyzeal.plugins.plugin_loader import PluginLoader
from pyzeal.pyzeal_logging.log_manager import LogManager
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger
from pyzeal.pyzeal_types.init_modes import InitModes
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.factories.algorithm_factory import AlgorithmFactory
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.factories.estimator_factory import EstimatorFactory
from pyzeal.utils.factories.settings_factory import SettingsServiceFactory
from pyzeal.utils.install_test_facade import InstallTestingHandlerFacade
from pyzeal.utils.install_test_handler import InstallTestingHandler
from pyzeal.utils.service_locator import ServiceLocator


class PyZEALInitializationHandler:
    "Static initialization handler for PyZEAL services."

    # the module level logger
    _logger: Optional[PyZEALLogger] = None

    # protect from multiple initializations
    initialized = False

    @staticmethod
    def initPyZEALServices(mode: InitModes = InitModes.SCRIPT) -> None:
        """
        Register all relevant services with the ServiceLocator and initialize
        the available plugins.
        """
        # check for re-initialization
        if PyZEALInitializationHandler.initialized:
            if PyZEALInitializationHandler._logger is None:
                return
            PyZEALInitializationHandler._logger.warning(
                "re-initialization attempt with mode %s detected - skipped!",
                mode.name if mode.name else "[unknown]",
            )
            return

        # register settings service first so we can initialize logging
        if mode in (InitModes.CLI | InitModes.SCRIPT):
            ServiceLocator.registerAsTransient(
                SettingsService, SettingsServiceFactory.getConcreteSettings
            )
        PyZEALInitializationHandler._logger = LogManager.initLogger(
            __name__.rsplit(".", maxsplit=1)[-1],
            ServiceLocator.tryResolve(SettingsService).logLevel,
        )

        if mode in InitModes.SCRIPT:
            PyZEALInitializationHandler._logger.info(
                "initializing algorithms..."
            )
            ServiceLocator.registerAsTransient(
                FinderAlgorithm, AlgorithmFactory.getConcreteAlgorithm
            )
            PyZEALInitializationHandler._logger.info(
                "initializing containers..."
            )
            ServiceLocator.registerAsTransient(
                RootContainer, ContainerFactory.getConcreteContainer
            )
            PyZEALInitializationHandler._logger.info(
                "initializing estimators..."
            )
            ServiceLocator.registerAsTransient(
                ArgumentEstimator, EstimatorFactory.getConcreteEstimator
            )
        if mode in InitModes.CLI:
            PyZEALInitializationHandler._logger.info("initializing cli...")
            ServiceLocator.registerAsSingleton(
                PyZEALParserInterface, PyZEALParser()
            )
            ServiceLocator.registerAsTransient(
                CLIControllerFacade, CLIController
            )
            ServiceLocator.registerAsTransient(
                InstallTestingHandlerFacade, InstallTestingHandler
            )

        # plugins cannot be loaded in cli mode (plugins might be broken, ...)!
        if mode not in InitModes.CLI:
            PyZEALInitializationHandler._logger.info("loading plugins...")
            PluginLoader.loadPlugins()

        # initialization complete!
        PyZEALInitializationHandler.initialized = True
        PyZEALInitializationHandler._logger.info("initialization complete!")



"""
TODO.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, runtime_checkable

from pyzeal.cli.parse_results import (
    InstallTestingParseResults,
    PluginParseResults,
    SettingsParseResults,
)


@runtime_checkable
class CLIControllerFacade(Protocol):
    """
    Interface for a generic controller instance used with the `PyZEAL` cli.
    """

    def handleViewSubcommand(self, args: SettingsParseResults) -> None:
        """
        Check if the 'view' subcommand was selected and print current settings.

        :param args: Parsed settings values
        """
        ...

    def handleChangeSubcommand(self, args: SettingsParseResults) -> None:
        """
        Check if the 'change' subcommand was selected and change settings
        accordingly.

        :param args: Parsed settings values
        """
        ...

    def handlePluginSubcommand(self, args: PluginParseResults) -> None:
        """
        Check if the 'plugin' subcommand was selected and manipulate plugins
        accordingly.

        :param args: Parsed plugin request values
        """
        ...

    def handleTestingOption(self, args: InstallTestingParseResults) -> bool:
        """
        Check if the '--test' option was selected and start testing the local
        `PyZEAL` installation accordingly.

        :param args: parsed testing option
        :return: flag indicating if the testing option was given
        """
        ...



"""
Timing benchmark suite for the simple argument (with and without numerical
quadrature) based root finding algorithm.

Authors:\n
- Luca Wasmuth\n
- Philipp Schuette\n
"""

from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_helpers import simpleArgumentRootFinder
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.utils.service_locator import ServiceLocator

benchmarkFunctions = [
    "x^2-1",
    "x^2+1",
    "x^4-1",
    "x^5-4x+2",
    "sin(x)",
    "exp(x)",
]

# disable progress bar
settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)


class SimpleArgumentSuite:
    """
    Timing benchmarks for the simple argument principle based root finder.
    Includes simple polynomials as well as trigonometric and exponential
    functions.
    """

    def TimeSimpleArgumentSummation(self) -> None:
        """
        Runs the simple argument version of the rootfinding algorithm with
        summation based argument estimation and non-parallel.
        """
        for caseName in benchmarkFunctions:
            hrf = simpleArgumentRootFinder(
                caseName,
                parallel=False,
                estimatorType=EstimatorTypes.SUMMATION_ESTIMATOR,
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            hrf.calculateRoots(reRan, imRan, precision=(5, 5))

    def TimeSimpleArgumentQuadrature(self) -> None:
        """
        Analogous to `TimeSimpleArgumentSummation` but here the algorithm uses
        quadrature based argument estimation.
        """
        for caseName in benchmarkFunctions:
            hrf = simpleArgumentRootFinder(
                caseName,
                parallel=False,
                estimatorType=EstimatorTypes.QUADRATURE_ESTIMATOR,
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            hrf.calculateRoots(reRan, imRan, precision=(5, 5))



"""
This module contains the resources necessary for the tests of the PyZEAL
project. The resources are provided as a dictionary mapping function names to
triples of holomorphic target functions, their derivatives, and their roots in
a certain domain of the complex plane.

Authors:\n
- Luca Wasmuth\n
"""

from dataclasses import dataclass
from functools import partial
from typing import Dict, Tuple, cast

import numpy as np

from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.pyzeal_types.root_types import tHoloFunc, tVec
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.rounding_container import RoundingContainer
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.root_context import RootContext
from pyzeal.utils.service_locator import ServiceLocator

ServiceLocator.registerAsTransient(SettingsService, RAMSettingsService)


# dictionary mapping function names to functions, their derivatives, expected
# roots, and the real/imaginary ranges where the roots should be searched
@dataclass
class TestData:
    "Container holding test data sets."
    testFunc: tHoloFunc
    testFuncDerivative: tHoloFunc
    expectedRoots: Tuple[complex, ...]
    reRan: Tuple[float, float]
    imRan: Tuple[float, float]
    precision: Tuple[int, int]


def buildContextFromData(data: TestData) -> RootContext:
    """
    Convenience function building a root finding context from given test data.

    :param data: the data used for context building
    """
    # build a simple container for the context
    container = RoundingContainer(precision=data.precision)
    ContainerFactory.registerPreDefinedFilter(
        container, filterType=FilterTypes.FUNCTION_VALUE_ZERO
    )
    ContainerFactory.registerPreDefinedFilter(
        container, filterType=FilterTypes.ZERO_IN_BOUNDS
    )

    # assemble the context from the data
    deltaRe, deltaIm = data.precision
    reRan, imRan = data.reRan, data.imRan
    reRan = reRan[0] - 0.001, reRan[1] + 0.002
    imRan = imRan[0] - 0.003, imRan[1] + 0.004
    return RootContext(
        f=data.testFunc,
        df=data.testFuncDerivative,
        container=container,
        precision=(deltaRe + 1, deltaIm + 1),
        reRan=reRan,
        imRan=imRan,
    )


localTestFunctions: Dict[str, TestData] = {
    "x^2-1": TestData(
        testFunc=lambda x: cast(tVec, x**2 - 1),
        testFuncDerivative=lambda x: 2 * x,
        expectedRoots=(-1, 1),
        reRan=(-3, 2),
        imRan=(-4, 5),
        precision=(3, 3),
    ),
    "x^2+1": TestData(
        testFunc=lambda x: cast(tVec, x**2 + 1),
        testFuncDerivative=lambda x: 2 * x,
        expectedRoots=(1j, -1j),
        reRan=(-5, 5),
        imRan=(-3, 3),
        precision=(3, 3),
    ),
    "x^2": TestData(
        testFunc=lambda x: cast(tVec, x**2),
        testFuncDerivative=lambda x: cast(tVec, 2 * x),
        expectedRoots=(),
        reRan=(1, 2),
        imRan=(-1, 1),
        precision=(5, 5),
    ),
    "(2x/5)^2+1": TestData(
        lambda x: cast(tVec, (2 * x / 5) ** 2 + 1),
        lambda x: 8 * x / 25,
        (2.5j, -2.5j),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^4-1": TestData(
        lambda x: cast(tVec, x**4 - 1),
        lambda x: cast(tVec, 4 * x**3),
        (1, -1, 1j, -1j),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^3+x^2+x+1": TestData(
        lambda x: cast(tVec, 1 * x**3 + x**2 + x + 1),
        lambda x: cast(tVec, 3 * x**2 + 2 * x + 1),
        (-1, 1j, -1j),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^2+26.01": TestData(
        lambda x: cast(tVec, x**2 + 26.01),
        lambda x: 2 * x,
        (),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^4-6.25x+9": TestData(
        lambda x: cast(
            tVec, (x - np.sqrt(2)) * (x + np.sqrt(2)) * (x - 1.5) * (x + 1.5)
        ),
        lambda x: cast(tVec, 4 * x**3 - 8.5 * x),
        (np.sqrt(2), -np.sqrt(2), 1.5, -1.5),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^5-4x+2": TestData(
        testFunc=lambda x: cast(tVec, 1 * x**5 - 4 * x + 2),
        testFuncDerivative=lambda x: cast(tVec, 5 * x**4 - 4),
        expectedRoots=(
            0.508499484,
            -1.51851215,
            1.2435963,
            -0.116791 - 1.43844769j,
            -0.116791 + 1.43844769j,
        ),
        reRan=(-7.1, 6.2),
        imRan=(-9.3, 8.4),
        precision=(3, 3),
    ),
    "x^3-0.01x": TestData(
        lambda x: cast(tVec, 1 * (x - 0.1) * (x + 0.1) * x),
        lambda x: cast(tVec, 3 * x**2 - 0.01),
        (0.1, 0, -0.1),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^30": TestData(
        lambda x: x**30,
        lambda x: cast(tVec, 30 * x**29),
        (0,),
        (-5.009, 5.01),
        (-5.02, 5.03),
        precision=(3, 3),
    ),
    "x^50": TestData(
        lambda x: x**50,
        lambda x: cast(tVec, 50 * x**49),
        (0,),
        (-5.009, 5.018),
        (-5.027, 5.036),
        precision=(3, 3),
    ),
    "x^100": TestData(
        lambda x: x**100,
        lambda x: cast(tVec, 100 * x**99),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "1e6 * x^100": TestData(
        lambda x: cast(tVec, 1e6 * x**100),
        lambda x: cast(tVec, 1e8 * x**99),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "x^2-25": TestData(
        lambda x: cast(tVec, 1 * (x - 5) * (x + 5)),
        lambda x: 2 * x,
        (5, -5),
        (-5.1, 5.2),
        (-5.3, 5.4),
        precision=(3, 3),
    ),
    "x^2+(0.000024414 - i)x": TestData(
        lambda x: x * (x + 0.000024414 - 1j),
        lambda x: cast(tVec, 2 * x + 0.000024414 - 1j),
        (0, -0.000024414 + 1j),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "poly": TestData(
        testFunc=lambda x: cast(
            tVec,
            (x + 4.652) * (x + 0.333 + 1.5j) * (x + 1.5j) * x * (x - 4.768j),
        ),
        testFuncDerivative=lambda x: cast(
            tVec,
            (
                (x + 0.333 + 1.5j) * (x + 1.5j) * x * (x - 4.768j)
                + (x + 4.652) * (x + 1.5j) * x * (x - 4.768j)
                + (x + 4.652) * (x + 0.333 + 1.5j) * x * (x - 4.768j)
                + (x + 4.652) * (x + 0.333 + 1.5j) * (x + 1.5j) * (x - 4.768j)
                + (x + 4.652) * (x + 0.333 + 1.5j) * (x + 1.5j) * x
            ),
        ),
        expectedRoots=(-4.652, -0.333 - 1.5j, -1.5j, 0, 4.768j),
        reRan=(-6.2, 7.4),
        imRan=(-8.6, 9.8),
        precision=(3, 3),
    ),
    "sin(x)": TestData(
        np.sin,
        np.cos,
        (-1 * np.pi, 0, np.pi),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "exp(x)": TestData(np.exp, np.exp, (), (-5, 5), (-5, 5), precision=(3, 3)),
    "exp(x)-1": TestData(
        testFunc=lambda x: cast(tVec, np.exp(np.pi * x / 2) - 1),
        testFuncDerivative=lambda x: cast(
            tVec, np.pi * np.exp(np.pi * x / 2) / 2
        ),
        expectedRoots=(-4j, 0, 4j),
        reRan=(-5, 5),
        imRan=(-5, 5),
        precision=(3, 3),
    ),
    "tan(x/10)": TestData(
        lambda x: cast(tVec, np.tan(x / 10)),
        lambda x: cast(tVec, 1 / (10 * np.cos(x / 10) ** 2)),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "tan(x/100)": TestData(
        lambda x: cast(tVec, np.tan(x / 100)),
        lambda x: cast(tVec, 1 / (100 * np.cos(x / 100) ** 2)),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "sin x cos": TestData(
        lambda x: cast(tVec, 2 * np.sin(x) * np.cos(x)),
        lambda x: cast(tVec, 2 * np.cos(x) ** 2 - 2 * np.sin(x) ** 2),
        (
            0,
            np.pi,
            -np.pi,
            0.5 * np.pi,
            1.5 * np.pi,
            -0.5 * np.pi,
            -1.5 * np.pi,
        ),
        (-5, 5.01),
        (-5, 5.02),
        precision=(3, 3),
    ),
    "log and sinh composition": TestData(
        lambda x: cast(tVec, np.log(np.sinh(0.2 * x) ** 2 + 1)),
        lambda x: cast(
            tVec,
            0.4
            * np.sinh(0.2 * x)
            * np.cosh(0.2 * x)
            / (1 + np.sinh(0.2 * x) ** 2),
        ),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "square root of exp": TestData(
        lambda x: cast(tVec, np.emath.sqrt(np.exp(2 * np.pi * x / 4)) - 1),
        lambda x: cast(
            tVec, 0.25 * np.pi / np.emath.sqrt(np.exp(2 * np.pi * x / 4))
        ),
        (0, -4j, 4j),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "log(x^2+26)": TestData(
        lambda x: cast(tVec, np.log(x**2 + 26)),
        lambda x: cast(tVec, 2 * x / (x**2 + 26)),
        (-5j, 5j),
        (-5.01, 5.02),
        (-5.03, 5.04),
        precision=(3, 3),
    ),
    "20arctan(x/10)": TestData(
        lambda x: cast(tVec, 20 * np.arctan(x / 10)),
        lambda x: cast(tVec, 2 / (1 + (x / 10) ** 2)),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
    "exp(arctan(x/10)-1)": TestData(
        lambda x: cast(tVec, np.exp(np.arctan(x / 10)) - 1),
        lambda x: cast(
            tVec, 0.1 * np.exp(np.arctan(x / 10)) / (1 + (x / 10) ** 2)
        ),
        (0,),
        (-5, 5),
        (-5, 5),
        precision=(3, 3),
    ),
}


# wrap lambdas inside partial so multiprocessing works correctly
def f(name: str, x: tVec) -> tVec:
    """
    Evaluate the function `name` with argument `x`. Use with `partial` to
    get a function in `x`.

    :param name: Function name
    :param x: Point for function evaluation
    :return: `name` evaluated at `x`
    """
    return localTestFunctions[name].testFunc(x)


def df(name: str, x: tVec) -> tVec:
    """
    Evaluate the derivative of function `name` at `x`.

    :param name: Function name
    :param x: Point for function evaluation
    :return: Derivative of `name` evaluated at `x`
    """
    return localTestFunctions[name].testFuncDerivative(x)


testFunctions: Dict[str, TestData] = {}
for key, item in localTestFunctions.items():
    testFunctions[key] = TestData(
        testFunc=partial(f, key),
        testFuncDerivative=partial(df, key),
        expectedRoots=localTestFunctions[key].expectedRoots,
        reRan=localTestFunctions[key].reRan,
        imRan=localTestFunctions[key].imRan,
        precision=localTestFunctions[key].precision,
    )



"""
This module provides type aliases used in conjunction with multiprocessing.

Authors:\n
- Philipp Schuette\n
"""

from multiprocessing.managers import BaseManager
from typing import Callable, Protocol, Tuple

from pyzeal.utils.finder_progress import FinderProgressBar


# typed queues to be used for message passing with multiprocessing.Queue
class tQueue(Protocol):
    r"""
    Queue that stores tuples of the form (root: complex, order: int).
    """

    def get(self) -> Tuple[complex, int]:
        "Get first element of the queue."
        ...

    def put(self, item: Tuple[complex, int]) -> None:
        "Put element into the queue."
        ...

    def empty(self) -> bool:
        "Check if queue is empty."
        ...


# multiprocessing managers used to share access to progress bars
class FinderProgressManager(BaseManager):
    r"""
    Multiprocessing.BaseManager containing a rich.progress.Progress instance.
    """
    finderProgress: Callable[..., FinderProgressBar]



"""
Module init_modes.py from the package PyZEAL.
This module contains named constants which identify the different startup modes
of the PyZEAL project (i.e. cli, scripting, ...).

Authors:\n
- Philipp Schuette\n
"""

from enum import Flag, auto


class InitModes(Flag):
    "Enumeration containing the different startup modes for PyZEAL."
    CLI = auto()
    SCRIPT = auto()
    GUI = auto()



"""
Module finder_interface of the package pyzeal.

This module contains the base interface for all rootfinder implementations. Its
methods provide the main root finding API for end users. Internally its
implementations should build up an appropriate root finding context and then
employ a `FinderAlgorithm` to calculate the actual roots. Internal book keeping
is facilitated by the implementations of `RootContainer`.

Authors:\n
- Philipp Schuette\n
"""

from abc import ABC, abstractmethod
from typing import Optional, Tuple

from numpy import int32
from numpy.typing import NDArray

from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.pyzeal_types.root_types import tVec
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.initialization_handler import PyZEALInitializationHandler


class RootFinderInterface(ABC):
    """
    Interface for the main root finding API. After initializing a root finder
    for a given concrete problem methods for calculating roots as well as
    properties for retrieving roots and their orders are available.
    """

    PyZEALInitializationHandler.initPyZEALServices()

    @abstractmethod
    def calculateRoots(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        precision: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Abstract entry point for starting a root finding calculation in the
        rectangle `reRan x imRan` up to a number of `precision` significant
        digits in real and imaginary part.

        :param reRan: horizontal extend of the complex region to search in
        :param imRan: vertical extend of the complex region to search in
        :param precision: accuracy of the search in real and imaginary parts
        """

    @property
    @abstractmethod
    def roots(self) -> tVec:
        """
        Return the roots calculated with this root finder through previous
        calls to `calculateRoots()`.

        :return: the set of roots calculated by this finder
        """

    @property
    @abstractmethod
    def orders(self) -> NDArray[int32]:
        """
        Return the orders of the roots calculated with this finder. The output
        is parallel to the return value of the `roots` property.

        :return: the orders of the roots calculated by this finder
        """

    @property
    @abstractmethod
    def container(self) -> RootContainer:
        """
        Return the container instance managing the roots of this finder.

        :return: the container of this finder
        """

    def setRootFilter(
        self, filterType: FilterTypes, *, threshold: int = 3
    ) -> None:
        """
        Add a filter of type `filterType` to the set of filters which
        found roots are checked against.

        :param filterType: the type of root filter to add
        :param threshold: the threshold for function on root evaluation filters
        """
        ContainerFactory.registerPreDefinedFilter(
            self.container, filterType, threshold=threshold
        )



"""
This module provides a straight-forward implementation of the `SettingsSerivce`
based on a `json` serialization backend.

Authors:\n
- Philipp Schuette\n
"""

from os.path import dirname, join
from typing import Dict, Final, Tuple, Union

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.invalid_setting_exception import InvalidSettingException
from pyzeal.settings.json_core_settings import JSONCoreSettingsService
from pyzeal.settings.json_helper import JSONHelper
from pyzeal.settings.settings_service import SettingsService

# location where the default settings should be
DEFAULT_SETTINGS: Final[str] = join(dirname(__file__), "default_settings.json")
# default location where custom settings are saved
CUSTOM_SETTINGS: Final[str] = join(dirname(__file__), "custom_settings.json")


class JSONSettingsService(SettingsService):
    """
    This class provides a layer of abstraction for storage and retrieval of
    PyZEAL related settings using JSON for persistence. Any read and/or write
    access to settings must happen through a service like this one.
    """

    __slots__ = ("_coreSettings", "_level", "_verbose", "_precision")

    def __init__(self) -> None:
        """
        Create an instance of a new `SettingsService`. The basis for its
        properties are the currently persisted (user or default) settings.
        """
        self._coreSettings = JSONCoreSettingsService()
        currentSettings: Dict[str, Union[str, bool, Tuple[int, int]]] = {}
        # first load default settings (must always exist)...
        JSONHelper.loadSettingsFromFile(DEFAULT_SETTINGS, currentSettings)
        # ...then try to load custom settings (might not exist)
        JSONHelper.loadSettingsFromFile(CUSTOM_SETTINGS, currentSettings)

        # set default logging level
        for level in LogLevel:
            if level.name == currentSettings["logLevel"]:
                self._level = level

        # set default verbosity
        verbosity = currentSettings.get("verbose", None)
        if verbosity is not None:
            self._verbose = bool(verbosity)

        # set default precision
        precision = currentSettings.get("precision", None)
        if precision is not None:
            try:
                if isinstance(precision, tuple):
                    self._precision = (
                        precision[0],
                        precision[1],
                    )
            except ValueError:
                pass

        # check if required settings were loaded successfully
        for attr in self.__slots__:
            if not hasattr(self, attr):
                raise InvalidSettingException(attr[1:])

    def __str__(self) -> str:
        """
        A printable string representation of the currently active settings
        configuration.

        :return: string representation of current setting
        """
        return (
            f"Currently active settings configuration:\n{self._coreSettings}"
            f"-> default log level:   {self.logLevel.name}\n"
            f"-> default verbosity:   {self.verbose}\n"
            f"-> default precision:   {self.precision}"
        )

    # docstr-coverage:inherited
    @property
    def defaultContainer(self) -> ContainerTypes:
        return self._coreSettings.defaultContainer

    # docstr-coverage:inherited
    @defaultContainer.setter
    def defaultContainer(self, value: ContainerTypes) -> None:
        self._coreSettings.defaultContainer = value

    # docstr-coverage:inherited
    @property
    def defaultAlgorithm(self) -> AlgorithmTypes:
        return self._coreSettings.defaultAlgorithm

    # docstr-coverage:inherited
    @defaultAlgorithm.setter
    def defaultAlgorithm(self, value: AlgorithmTypes) -> None:
        self._coreSettings.defaultAlgorithm = value

    # docstr-coverage:inherited
    @property
    def defaultEstimator(self) -> EstimatorTypes:
        return self._coreSettings.defaultEstimator

    # docstr-coverage:inherited
    @defaultEstimator.setter
    def defaultEstimator(self, value: EstimatorTypes) -> None:
        self._coreSettings.defaultEstimator = value

    # docstr-coverage:inherited
    @property
    def logLevel(self) -> LogLevel:
        return self._level

    # docstr-coverage:inherited
    @logLevel.setter
    def logLevel(self, value: LogLevel) -> None:
        self._level = value
        JSONHelper.createOrUpdateSetting(CUSTOM_SETTINGS, "logLevel", value)

    # docstr-coverage:inherited
    @property
    def verbose(self) -> bool:
        return self._verbose

    # docstr-coverage:inherited
    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._verbose = value
        JSONHelper.createOrUpdateSetting(CUSTOM_SETTINGS, "verbose", value)

    # docstr-coverage:inherited
    @property
    def precision(self) -> Tuple[int, int]:
        return self._precision

    # docstr-coverage:inherited
    @precision.setter
    def precision(self, value: Tuple[int, int]) -> None:
        self._precision = value
        JSONHelper.createOrUpdateSetting(CUSTOM_SETTINGS, "precision", value)



"""
Module service_locator.py from the package PyZEAL.
This module provides a way to register and locate services used throughout
the project, to enable loading additional services for the plugin system.

Authors:\n
- Philipp Schuette\n
"""

from __future__ import annotations

from inspect import Parameter, signature
from typing import Callable, Dict, Type, TypeVar

from pyzeal.utils.configuration_exception import InvalidServiceConfiguration

# type variable for the various signatures of ServiceLocator
T = TypeVar("T")


class ServiceLocator:
    """
    Static locator class used to add and resolve services.
    """

    _transientServices: Dict[Type[object], Callable[..., object]] = {}

    _singletonServices: Dict[Type[object], object] = {}

    _sealed: bool = False

    @staticmethod
    def registerAsSingleton(
        serviceType: Type[T], instance: T
    ) -> Type[ServiceLocator]:
        """
        Register service `instance` as a singleton service, meaning only one
        instance exists.

        :param serviceType: Type of service
        :param instance: Service instance
        :raises ValueError: If the service locator is sealed, no new services
            can be registered.
        :raises InvalidServiceConfiguration: Given instance must implement the
            given type
        :return: Return the static `ServiceLocator` for method chaining
        """
        if ServiceLocator.isSealed():
            raise ValueError(
                f"cannot register singleton {serviceType} on sealed locator!"
            )
        if isinstance(instance, serviceType):
            ServiceLocator._singletonServices[serviceType] = instance
            return ServiceLocator
        raise InvalidServiceConfiguration(serviceType)

    @staticmethod
    def registerAsTransient(
        serviceType: Type[T], factory: Callable[..., T]
    ) -> Type[ServiceLocator]:
        """
        Register a transient service. Note that you MUST implement a (dummy)
        default constructor if you want to register a class without constructor
        as an instance factory and that class inherits from `typing.Protocol`.

        :param serviceType: Type of service to register.
        :param factory: Factory for the given service
        :raises ValueError: If the service locator is sealed, no new services
            can be registered.
        :return: Return the static `ServiceLocator` for method chaining
        """
        if ServiceLocator.isSealed():
            raise ValueError(
                f"cannot register transient {serviceType} on sealed locator!"
            )
        ServiceLocator._transientServices[serviceType] = factory
        return ServiceLocator

    @staticmethod
    def tryResolve(serviceType: Type[T], **kwargs: object) -> T:
        """
        Try to resolve the requested service type by first searching registered
        singleton and then registered transient configurations.

        :param serviceType: Type of service to resolve
        :raises InvalidServiceConfiguration: If the given `serviceType` can
            not be resolved, an exception is raised
        :return: An instance of the given service. If the service is
            transient, the factory is called with the parameters given
            by `**kwargs`.
        """
        # try to resolve as singleton
        instance = ServiceLocator._singletonServices.get(serviceType, None)
        if isinstance(instance, serviceType):
            return instance

        # try to resolve as transient
        factory = ServiceLocator._transientServices.get(serviceType, None)
        if factory is None:
            raise InvalidServiceConfiguration(serviceType)
        sig = signature(factory)
        arguments: Dict[str, object] = {}

        # try to instantiate an instance from the factory and given kwargs or
        # default parameters of the factory - additional kwargs are ignored
        for param in sig.parameters:
            try:
                arguments[param] = kwargs[param]
            except KeyError:
                if (arg := sig.parameters[param].default) != Parameter.empty:
                    arguments[param] = arg
                else:
                    arguments[param] = ServiceLocator.tryResolve(
                        sig.parameters[param].annotation
                    )
        instance = factory(**arguments)
        if isinstance(instance, serviceType):
            return instance

        # resolving failed
        raise InvalidServiceConfiguration(serviceType)

    @staticmethod
    def seal() -> None:
        """
        Seal the service locator to prevent additional services from
        being registered.

        :raises ValueError: Raises an exception if the service locator
            has already been sealed.
        """
        if ServiceLocator._sealed:
            raise ValueError("cannot re-seal an already sealed locator!")
        ServiceLocator._sealed = True

    @staticmethod
    def isSealed() -> bool:
        """
        Return `True` if the service locator is sealed.

        :return: `True` if the service locator is sealed.
        """
        return ServiceLocator._sealed

    @staticmethod
    def clearConfigurations() -> None:
        """
        Clear all service locator configurations, unsealing the locator in the
        process.
        """
        ServiceLocator._transientServices.clear()
        ServiceLocator._singletonServices.clear()
        ServiceLocator._sealed = False



"""
Class FinderProgressBar from the package pyzeal_utils.
This module defines a progress bar used for command line display of the
progress a running root finding algorithm has made.

Authors:\n
- Philipp Schuette\n
"""

from os import getpid

from rich.progress import Progress, SpinnerColumn, TaskID, TimeElapsedColumn


class FinderProgressBar(Progress):
    """
    This class is a very specialized subtype of the general `rich.Progress`
    type representing a very customizable command line displayed progress bar.
    """

    def __init__(self) -> None:
        "Initialize a root finder specific progress bar with fixed defaults."
        super().__init__(
            *Progress.get_default_columns()[:],
            SpinnerColumn(),
            "Elapsed:",
            TimeElapsedColumn(),
            transient=False,
            refresh_per_second=3,
            speed_estimate_period=10,
            expand=True,
        )

    def addTask(self, total: float) -> TaskID:
        """
        Override the `add_task()` method of generic progress bars by allowing
        only the total amount of work to be adjusted and prescribing a default
        task description.

        :param total: the total amount of work contained in the added task
        """
        return super().add_task(
            description=(f"[magenta]{getpid()}: " + "[g]getting roots..."),
            total=total,
        )



"""
This module provides an argument estimator based on numerical integration
using Romberg quadrature.

Authors:\n
- Philipp Schuette
"""

from typing import Tuple

import numpy as np
from scipy.integrate import romb  # type: ignore

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.estimators.constants import (
    EXP_SAMPLE_POINTS,
    MAX_SAMPLE_POINTS,
)
from pyzeal.algorithms.estimators.estimator_cache import EstimatorCache
from pyzeal.pyzeal_types.root_types import tVec
from pyzeal.utils.root_context import RootContext


class QuadratureEstimator(ArgumentEstimator):
    """
    This class implements an argument estimator using numerical quadrature
    to integrate the logarithmic derivative.
    """

    __slots__ = ("_cache",)

    def __init__(self, *, cache: EstimatorCache) -> None:
        """
        Initialize a `QuadratureEstimator`.

        :param cache: Cache to store intermediate values in.
        """
        self._cache = cache

    def calcMomentAlongLine(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
        context: RootContext,
    ) -> complex:
        """
        Calculate the `order`-th moment of the logarithmic derivative along
        the line given by `zStart` and `zEnd`.

        :param order: Moment to compute
        :param zStart: Start z-value
        :param zEnd: End z-value
        :param context: `RootContext` containing the necessary information
        :raises ValueError: An error is raised when no derivative is supplied,
            as the `QuadratureEstimator` does not support derivative-free
            argument estimation.
        :return: The moment as calculated along the given line.
        """
        if context.df is None:
            raise ValueError(
                "derivative required for quadrature-based argument estimation!"
            )

        samplePts = EXP_SAMPLE_POINTS
        zArr, funcArr = self.genFuncArr(
            zStart, zEnd, context, 2**samplePts + 1
        )
        funcArr = context.df(zArr) * zArr**order / funcArr
        distance = abs(zEnd - zStart)

        realResult = romb(np.real(funcArr), distance / (2**samplePts))
        imagResult = romb(np.imag(funcArr), distance / (2**samplePts))
        while samplePts <= MAX_SAMPLE_POINTS:
            # for expensive target functions merging of arrays with function
            # values is preferable over recalculation - for inexpensive targets
            # the converse is true:
            merge = samplePts >= EXP_SAMPLE_POINTS + 2
            if merge:
                zStart = (zArr[0] + zArr[1]) / 2
                zEnd = (zArr[-1] + zArr[-2]) / 2
                newZArr, newFuncArr = self.genFuncArr(
                    zStart, zEnd, context, 2**samplePts
                )
                newFuncArr = (
                    context.df(newZArr) * newZArr**order / newFuncArr
                )
                zArr, funcArr = QuadratureEstimator.mergeArrays(
                    (zArr, funcArr), (newZArr, newFuncArr)
                )

                samplePts += 1
            else:
                samplePts += 1
                zArr, funcArr = self.genFuncArr(
                    zStart, zEnd, context, 2**samplePts + 1
                )
                funcArr = context.df(zArr) * zArr**order / funcArr

            newRealResult = romb(np.real(funcArr), distance / (2**samplePts))
            newImagResult = romb(np.imag(funcArr), distance / (2**samplePts))
            if (
                abs(newRealResult - realResult) < 1e-3
                and abs(newImagResult - imagResult) < 1e-3
            ):
                break
            realResult = newRealResult
            imagResult = newImagResult

        # result (divided by 1j) is only necessarily real if order=0!
        return complex(
            (zEnd - zStart) * (-1j * realResult + imagResult) / distance
        )

    # docstr-coverage:inherited
    @property
    def cache(self) -> EstimatorCache:
        return self._cache

    @staticmethod
    def mergeArrays(
        oldArrays: Tuple[tVec, tVec], newArrays: Tuple[tVec, tVec]
    ) -> Tuple[tVec, tVec]:
        """
        TODO.
        """
        zArr, funcArr = oldArrays
        newZArr, newFuncArr = newArrays
        zArr = np.append(np.vstack((zArr[:-1], newZArr)).ravel("F"), zArr[-1])
        funcArr = np.append(
            np.vstack((funcArr[:-1], newFuncArr)).ravel("F"), funcArr[-1]
        )
        return zArr, funcArr



"""
Class JSONCoreSettingsService from the module `pyzeal.settings`.

This module provides a straight-forward implementation of the
`CoreSettingsSerivce` based on a `json` serialization backend.

Authors:\n
- Philipp Schuette\n
"""

from os.path import dirname, join
from typing import Dict, Final

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.core_settings_service import CoreSettingsService
from pyzeal.settings.invalid_setting_exception import InvalidSettingException
from pyzeal.settings.json_helper import JSONHelper

# location where the default settings should be
DEFAULT_SETTINGS: Final[str] = join(dirname(__file__), "default_settings.json")
# default location where custom settings are saved
CUSTOM_SETTINGS: Final[str] = join(dirname(__file__), "custom_settings.json")


class JSONCoreSettingsService(CoreSettingsService):
    """
    This class provides a layer of abstraction for storage and retrieval of
    PyZEAL core settings using JSON for persistence. Any read and/or write
    access to (core) settings must happen through a service like this one.
    """

    __slots__ = ("_container", "_algorithm", "_estimator")

    def __init__(self) -> None:
        """
        Create an instance of a new `CoreSettingsService`. The basis for its
        properties are the currently persisted (user or default) settings.
        """
        currentSettings: Dict[str, str] = {}
        # first load default settings (must always exist)...
        JSONHelper.loadCoreSettingsFromFile(DEFAULT_SETTINGS, currentSettings)
        # ...then try to load custom settings (might not exist)
        JSONHelper.loadCoreSettingsFromFile(CUSTOM_SETTINGS, currentSettings)

        # set default container
        for container in ContainerTypes:
            if container.value == currentSettings["defaultContainer"]:
                self._container = container
                break

        # set default algorithm
        for algorithm in AlgorithmTypes:
            if algorithm.value == currentSettings["defaultAlgorithm"]:
                self._algorithm = algorithm
                break

        # set default estimator
        for estimator in EstimatorTypes:
            if estimator.value == currentSettings["defaultEstimator"]:
                self._estimator = estimator
                break

        # check if required defaults were loaded successfully
        for attr in self.__slots__:
            if not hasattr(self, attr):
                raise InvalidSettingException(attr[1:])

    def __str__(self) -> str:
        """
        A printable string representation of the currently active core settings
        configuration.

        :return: string representation of current core setting
        """
        return (
            f"-> default container:   {self.defaultContainer.value}\n"
            f"-> default algorithm:   {self.defaultAlgorithm.value}\n"
            f"-> default estimator:   {self.defaultEstimator.value}\n"
        )

    # docstr-coverage:inherited
    @property
    def defaultContainer(self) -> ContainerTypes:
        return self._container

    # docstr-coverage:inherited
    @defaultContainer.setter
    def defaultContainer(self, value: ContainerTypes) -> None:
        self._container = value
        JSONHelper.createOrUpdateCoreSetting(
            CUSTOM_SETTINGS, "defaultContainer", value
        )

    # docstr-coverage:inherited
    @property
    def defaultAlgorithm(self) -> AlgorithmTypes:
        return self._algorithm

    # docstr-coverage:inherited
    @defaultAlgorithm.setter
    def defaultAlgorithm(self, value: AlgorithmTypes) -> None:
        self._algorithm = value
        JSONHelper.createOrUpdateCoreSetting(
            CUSTOM_SETTINGS, "defaultAlgorithm", value
        )

    # docstr-coverage:inherited
    @property
    def defaultEstimator(self) -> EstimatorTypes:
        return self._estimator

    # docstr-coverage:inherited
    @defaultEstimator.setter
    def defaultEstimator(self, value: EstimatorTypes) -> None:
        self._estimator = value
        JSONHelper.createOrUpdateCoreSetting(
            CUSTOM_SETTINGS, "defaultEstimator", value
        )



"""
The `algorithms` package contains the core root finding algorithm of `PyZEAL`.

A `FinderAlgorithm` accepts a request to calculate roots in some given
`RootContext`. Under the hood it combines some concrete refinement strategy
with a concrete `ArgumentEstimator`. The algorithm produces its results by
using the latter's results as an informational basis for the former.
"""

from pyzeal.algorithms.finder_algorithm import FinderAlgorithm

__all__ = [
    "FinderAlgorithm",
]



"""
This module provides named constants used to identify concrete settings
services for the selection of a concrete settings model.

Authors:\n
- Philipp Schuette\n
"""

from enum import Enum


class SettingsServicesTypes(Enum):
    "Enumeration containing named constants identifying available settings."
    DEFAULT = 0
    JSON_SETTINGS = 1



"""
Module log_manager.py of the pyzeal_logging package.
This module contains the basic logging setup to be used with all PyZEAL related
modules.

Authors:\n
- Philipp Schuette\n
"""

import logging
import os
from datetime import datetime
from os.path import join
from typing import Final, Tuple

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger


class LogManager:
    """
    This class handles basic logging functionalities to be used by
    PyZEAL-related modules.
    """

    DATE: Final[datetime] = datetime.now()
    LOG_DIR: Final[str] = "./logs/"
    LOG_FORMAT: Final[Tuple[str, str]] = (
        "[%(asctime)s:%(msecs)03d][%(name)s] %(message)s [%(levelname)s]",
        "%H:%M:%S",
    )
    LOG_PREFIX: Final[str] = "pyzeal_"
    LOG_EXT: Final[str] = ".log"

    @staticmethod
    def initLogger(logName: str, logLevel: LogLevel) -> PyZEALLogger:
        """
        Initialize a module-level logger for the module 'modName'. Default
        logging level is retrieved from `SettingsService`. All logs are stored
        in a ./logs directory in a file pyzeal_<datetime.now>.log, where the
        date is determined at the start of every new session (upon first
        creation of a module logger within a session).

        :param logName: the name of the logger, should equal the module name
        :return: the (module-level) logger
        """
        if not os.path.exists(LogManager.LOG_DIR):
            os.mkdir(LogManager.LOG_DIR)

        logger = logging.getLogger(logName)
        logger.setLevel(logLevel.value)

        if not logger.hasHandlers():
            fileName = LogManager.buildFileName()
            fHandler = logging.FileHandler(join(LogManager.LOG_DIR, fileName))
            formatter = logging.Formatter(*LogManager.LOG_FORMAT)
            fHandler.setFormatter(formatter)
            logger.addHandler(fHandler)

        return logger

    @staticmethod
    def buildFileName() -> str:
        """
        Build a name for a (new or existing) log file based on the PyZEAL
        logging naming convention.

        :return: convention compliant log file name
        """
        date = LogManager.DATE
        fileName = (
            f"{str(date.day).zfill(2)}{str(date.month).zfill(2)}"
            f"{date.year}{str(date.hour).zfill(2)}"
            f"{str(date.minute).zfill(2)}{str(date.second).zfill(2)}"
        )
        return LogManager.LOG_PREFIX + fileName + LogManager.LOG_EXT



"""
This module provides named constants used to identify concrete root finding
algorithms in any instance where such an algorithm needs to be specified.

Authors:\n
- Philipp Schuette\n
"""

from enum import Enum


class AlgorithmTypes(Enum):
    "Enumeration containing named constants identifying available algorithms."
    NEWTON_GRID = "NewtonGrid"
    SIMPLE_ARGUMENT = "SimpleArgument"
    SIMPLE_ARGUMENT_NEWTON = "SimpleArgumentNewton"
    ASSOCIATED_POLYNOMIAL = "AssociatedPolynomial"
    DEFAULT = "DefaultAlgorithm"



"""
This module provides a static factory class which maps the containers available
in the `ContainerTypes` enumeration to appropriate implementations of the
interface `RootContainer`.

Authors:\n
- Philipp Schuette\n
"""

from functools import partial
from typing import Callable, Optional, Tuple, cast

import numpy as np

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.log_manager import LogManager
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.pyzeal_types.parallel_types import tQueue
from pyzeal.pyzeal_types.root_types import tRoot
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.plain_container import PlainContainer
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.containers.rounding_container import RoundingContainer
from pyzeal.utils.filter_context import FilterContext
from pyzeal.utils.service_locator import ServiceLocator


class ContainerFactory:
    "Static factory class used to create instances of root containers."

    # the module level logger
    _logger: Optional[PyZEALLogger] = None

    @staticmethod
    def _func_value_zero(
        threshold: int, root: tRoot, context: FilterContext
    ) -> bool:
        """
        Filter predicate to determine if a possible root has a function
        value sufficiently close to zero

        :param threshold: A function is considered zero if its absolute value
            is below 10^-(threshold)
        :param root: Root candidate
        :param context: Context containing information about the search
        :return: `True` if the function is close enough to zero
        """
        return cast(
            bool,
            np.abs(context.f(np.array([root[0]], dtype=np.complex128)))
            < 10 ** (-threshold),
        )

    @staticmethod
    def _zero_in_bounds(root: tRoot, context: FilterContext) -> bool:
        """
        Filter predicate to determine if `root` is in bounds for `context`.

        :param root: Root to filter
        :param context: Context containing information about the search
        :return: `True` if root is in bounds
        """
        return (
            context.reRan[0] <= root[0].real <= context.reRan[1]
            and context.imRan[0] <= root[0].imag <= context.imRan[1]
        )

    @staticmethod
    def getConcreteContainer(
        containerType: ContainerTypes = ContainerTypes.DEFAULT,
        *,
        precision: Tuple[int, int] = (3, 3),
        queue: Optional[tQueue] = None,
    ) -> RootContainer:
        """
        Initialize and return a root container instance based on the given type
        `algoType` of container and a given accuracy in real and imaginary
        parts with default filters.

        :param containerType: type of container to construct
        :param precision: the accuracy of the given container
        :param queue: an existing queue instance as base for a plain container
        :return: a concrete `RootContainer` instance
        """
        if ContainerFactory._logger is None:
            ContainerFactory._logger = LogManager.initLogger(
                __name__.rsplit(".", maxsplit=1)[-1],
                ServiceLocator.tryResolve(SettingsService).logLevel,
            )
        if containerType == ContainerTypes.ROUNDING_CONTAINER:
            ContainerFactory._logger.debug(
                "requested a new rounding container..."
            )
            return RoundingContainer(precision)
        if containerType == ContainerTypes.PLAIN_CONTAINER:
            ContainerFactory._logger.debug(
                "requested a new plain container..."
            )
            return PlainContainer(queue)

        # return the current default container
        ContainerFactory._logger.debug("requested a new default container...")
        settings = ServiceLocator.tryResolve(SettingsService)
        container = ContainerFactory.getConcreteContainer(
            settings.defaultContainer,
            precision=precision,
            queue=queue,
        )
        return container

    @staticmethod
    def registerPreDefinedFilter(
        container: RootContainer,
        filterType: FilterTypes,
        *,
        threshold: int = 3,
    ) -> None:
        """
        Given a container instance this method adds a filter to its addRoot()
        method. The filter checks if the root satisfies some given condition in
        a concrete `FilterContext` and discards the root otherwise.

        :param container: the container which receives a new filter
        :param filterType: one of the predefined filters
        :param threshold: the threshold for evaluation of `f` on roots
        :return: the enhanced container
        """
        predicate: Callable[[tRoot, FilterContext], bool]
        if filterType == FilterTypes.FUNCTION_VALUE_ZERO:
            predicate = partial(ContainerFactory._func_value_zero, threshold)
        elif filterType == FilterTypes.ZERO_IN_BOUNDS:
            predicate = ContainerFactory._zero_in_bounds

        container.registerFilter(predicate, filterType.value)
        if ContainerFactory._logger is not None:
            ContainerFactory._logger.debug(
                "registered a new %s filter!", filterType.value
            )

    @staticmethod
    def setLevel(level: LogLevel) -> None:
        """
        Set the log level.

        :param level: the new log level
        """
        if ContainerFactory._logger is not None:
            ContainerFactory._logger.setLevel(level=level.value)

    @staticmethod
    def registerDefaultFilters(container: RootContainer) -> None:
        """
        Register a set of default filters for the given (rounding) container.

        :param container: Container to enable filters for
        """
        if isinstance(container, RoundingContainer):
            for filterType in [
                FilterTypes.FUNCTION_VALUE_ZERO,
                FilterTypes.ZERO_IN_BOUNDS,
            ]:
                ContainerFactory.registerPreDefinedFilter(
                    container, filterType=filterType
                )



"""
Implementation RoundingContainer of the RootContainer protocol from the
pyzeal_utils package.
The concrete container class implemented here automatically rounds added roots
to a fixed number of decimal places.

Authors:\n
- Philipp Schuette\n
"""

from logging import INFO
from typing import Dict, Optional, Set, Tuple

import numpy as np
from numpy.typing import NDArray

from pyzeal.pyzeal_types.root_types import tRoot, tVec
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.filter_context import FilterContext, tRootFilter
from pyzeal.utils.service_locator import ServiceLocator


class RoundingContainer(RootContainer):
    """
    This simple container implementation rounds every added root to a given
    number of decimal places and simple ignores any attempts to add additional
    roots which coincide with a previously added root after rounding. The
    multiplicity is not taken into account when comparing old and new roots.
    Changing the desired accuracy automatically removes all calculated roots
    to preserve consistency.
    """

    __slots__ = ("precision", "rootSet", "filters")

    def __init__(self, precision: Optional[Tuple[int, int]]) -> None:
        """
        Initialize a rounding RootContainer. If no precision is given,
        default precision is used.

        :param precision: expected accuracy of roots to be added
        """
        self.precision = (
            precision or ServiceLocator.tryResolve(SettingsService).precision
        )
        self.rootSet: Set[tRoot] = set()
        self.filters: Dict[str, tRootFilter] = {}
        self.logger.info("initialized a new rounding root container")

    def addRoot(self, root: tRoot, context: FilterContext) -> None:
        """
        Add a new root with given accuracy to the container. If the accuracy
        differs from the accuracy of roots already added then all previous
        roots are removed.

        :param root: the root to be added to the container
        :param context: the context of the new root, required for filtering
        """
        for filterPredicate in self.filters.values():
            if not filterPredicate(root, context):
                self.logger.debug(
                    "root %f+%fi was rejected by container filter!",
                    root[0].real,
                    root[0].imag,
                )
                return
        if context.precision != self.precision:
            self.clear()
            self.logger.debug(
                "new accuracy detected - rounding container cleared!"
            )
            self.precision = context.precision
        self.logger.debug(
            "attempting to add new root %f + %fi to rounding container!",
            root[0].real,
            root[0].imag,
        )
        roundedRoot = RoundingContainer.roundRoot(root, self.precision)
        if self.logger.isEnabledFor(INFO) and roundedRoot in self.rootSet:
            self.logger.info("duplicate root discarded by rounding container!")
            return
        self.logger.info(
            "new root %f + %fi added to rounding container",
            roundedRoot[0].real,
            roundedRoot[0].imag,
        )
        self.rootSet.add(roundedRoot)

    def removeRoot(self, root: tRoot) -> bool:
        """
        Remove a given root from the container. Return value indicates success.

        :param root: the root to be removed from the container
        :return: a boolean flag indicating if a removal happened
        """
        try:
            self.rootSet.remove(
                RoundingContainer.roundRoot(root, self.precision)
            )
            self.logger.debug(
                "removed root %f+%fi from rounding container!",
                root[0].real,
                root[0].imag,
            )
            return True
        except KeyError:
            self.logger.debug(
                "failed to remove root %f+%fi from rounding container!",
                root[0].real,
                root[0].imag,
            )
            return False

    def getRoots(self) -> tVec:
        """
        Returns all roots currently held in this container as a vector.

        :return: a vector of complex roots
        """
        result = np.empty((len(self.rootSet)), dtype=np.complex128)
        for i, root in enumerate(self.rootSet):
            result[i] = root[0]
        return result

    def getRootOrders(self) -> NDArray[np.int32]:
        """
        Returns the orders of all roots currently held in this container as a
        vector which is parallel to the vector returned by `getRoots`.

        :return: a vector of integer root orders (multiplicities)
        """
        result = np.empty((len(self.rootSet)), dtype=np.int32)
        for i, root in enumerate(self.rootSet):
            result[i] = root[1]
        return result

    def clear(self) -> None:
        "Clear the container by removing all roots."
        self.rootSet.clear()

    @staticmethod
    def roundRoot(root: tRoot, precision: Tuple[int, int]) -> tRoot:
        """
        Round a given root to a given number of decimal places.

        :param root: a root to be rounded
        :param precision: the number of decimal places to round to
        :return: the rounded root (multiplicity stays constant)
        """
        x, y = root[0].real, root[0].imag
        return complex(round(x, precision[0]), round(y, precision[1])), root[1]

    def registerFilter(self, filterPredicate: tRootFilter, key: str) -> None:
        """
        Register a new filter to check possible roots against

        :param filterPredicate: New filter to register
        :param key: A key to identify this filter
        """
        self.filters[key] = filterPredicate

    def unregisterFilter(self, key: str) -> None:
        """
        Remove the filter identified by `key`.

        :param key: Filter key
        """
        try:
            self.filters.pop(key)
            self.logger.debug("removed filter %s", key)
        except KeyError:
            self.logger.info("tried to remove non-existent filter %s!", key)



"""
Implementation PlainContainer of the RootContainer protocol from the
pyzeal_utils package.
The concrete container class implemented here simply adds roots to an internal
buffer without any further action. Its main use is the collection of potential
roots in subprocesses during parallel calculations.

Authors:\n
- Philipp Schuette\n
"""

from multiprocessing import Manager
from typing import List, Optional, cast

import numpy as np
from numpy.typing import NDArray

from pyzeal.pyzeal_types.parallel_types import tQueue
from pyzeal.pyzeal_types.root_types import tRoot, tVec
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.filter_context import FilterContext, tRootFilter


class PlainContainer(RootContainer):
    """
    Minimal container implementation. Simply adds roots to an internal buffer
    of type `tQueue` without any further action.

    This class is mainly designed for application with a `queue` instance
    obtained from a `multiprocessing.Manager`. With this particular choice of
    root buffer it can serve as shared memory for several distinct processes
    each running their own instance of a `FinderAlgorithm`.
    """

    __slots__ = ("rootBuffer", "_roots")

    def __init__(self, queue: Optional[tQueue]) -> None:
        """
        Initialize a new PlainContainer.

        :param queue: the internal buffer used for root storage
        """
        self.rootBuffer = cast(tQueue, queue or Manager().Queue())
        self._roots: List[tRoot] = []
        self.logger.info("initialized a new plain root container")

    def addRoot(self, root: tRoot, context: FilterContext) -> None:
        """
        Add a new root to the internal buffer, ignoring the filter context.

        :param root: the root to be added to the container
        :param context: the context of the new root, ignored
        """
        self.logger.debug(
            "adding new root %f + %fi to plain container!",
            root[0].real,
            root[0].imag,
        )
        self.rootBuffer.put(root)

    def removeRoot(self, root: tRoot) -> bool:
        """
        Roots cannot be removed from a plain container.

        :param root: the root to be removed from the container
        :return: always returns False
        """
        return False

    def getRoots(self) -> tVec:
        """
        Returns all roots currently held in this container as a vector.

        :return: a vector of complex roots
        """
        self._transferRootBuffer()
        return np.array([root for root, _ in self._roots], dtype=np.complex128)

    def getRootOrders(self) -> NDArray[np.int32]:
        """
        Returns the orders of all roots currently held in this container as a
        vector which is parallel to the vector returned by `getRoots`.

        :return: a vector of integer root orders (multiplicities)
        """
        self._transferRootBuffer()
        return np.array([order for _, order in self._roots], dtype=np.int32)

    def _transferRootBuffer(self) -> None:
        """
        Transfer roots from the root buffer to an internal list of roots.
        """
        while not self.rootBuffer.empty():
            self._roots.append(self.rootBuffer.get())

    def clear(self) -> None:
        "Plain containers cannot be cleared."
        raise NotImplementedError("cannot clear plain containers!")

    def registerFilter(self, filterPredicate: tRootFilter, key: str) -> None:
        """
        Cannot register filters with an instance of `PlainContainer`.

        :param filterPredicate: New filter to register
        :param key: A key to identify this filter
        """
        raise NotImplementedError("plain containers do not support filtering!")

    def unregisterFilter(self, key: str) -> None:
        """
        Cannot unregister filters with an instance of `PlainContainer`.

        :param key: Filter key
        """
        raise NotImplementedError("plain containers do not support filtering!")



"""
This module provides named constants used to identify concrete root container
implementations in any instance where such a container is required.

Authors:\n
- Philipp Schuette\n
"""

from enum import Enum


class ContainerTypes(Enum):
    "Enumeration containing named constants identifying available containers."
    ROUNDING_CONTAINER = "RoundingContainer"
    PLAIN_CONTAINER = "PlainContainer"
    DEFAULT = "DefaultContainer"



"""
Class FilterContext from the package pyzeal_util.
This module defines a data container that holds the information necessary for
a generic root filter predicate to function.

Authors:\n
- Philipp Schuette\n
"""

from dataclasses import dataclass
from typing import Callable, Tuple

from pyzeal.pyzeal_types.root_types import tHoloFunc, tRoot


@dataclass(frozen=True)
class FilterContext:
    """
    Container for the data context of a root filter predicate. The container is
    read-only.
    """

    f: tHoloFunc
    reRan: Tuple[float, float]
    imRan: Tuple[float, float]
    precision: Tuple[int, int]
    threshold: int = 3


# type used for filter predicates that filter roots upon container insertion
tRootFilter = Callable[[tRoot, FilterContext], bool]



"""
TODO.

Authors:\n
- Philipp\
"""

from typing import Protocol, Union, runtime_checkable

from pytest import ExitCode


@runtime_checkable
class InstallTestingHandlerFacade(Protocol):
    """
    An interface for a simple helper that handles testing of local
    installations.
    """

    def testModule(self, module: str) -> Union[int, ExitCode]:
        """
        Test a given module of the local `PyZEAL` installation.

        :param module: path to the module under test
        :return: exit status of the test conducted
        """
        ...



"""
This module contains tests of the SIMPLE_ARGUMENT_NEWTON implementation
of the root finding algorithm interface.

Authors:\n
- Luca Wasmuth\n
"""

import numpy as np
import pytest

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.simple_holo_newton import SimpleArgumentNewtonAlgorithm
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_test_cases import (
    buildContextFromData,
    testFunctions,
)
from pyzeal.tests.resources.utils import rootsMatchClosely
from pyzeal.utils.factories.estimator_factory import EstimatorFactory
from pyzeal.utils.service_locator import ServiceLocator

# disable progress bar by default for tests
settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)
ServiceLocator.registerAsTransient(
    ArgumentEstimator, EstimatorFactory.getConcreteEstimator
)

# some test functions do not work due to z-refinement limitations
KNOWN_FAILURES = ["x^100", "1e6 * x^100"]


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("testName", sorted(testFunctions.keys()))
@pytest.mark.parametrize(
    "estimator",
    [EstimatorTypes.SUMMATION_ESTIMATOR, EstimatorTypes.QUADRATURE_ESTIMATOR],
)
def testSimpleArgumentNewton(testName: str, estimator: EstimatorTypes) -> None:
    """
    Test the SIMPLE_ARGUMENT_NEWTON RootFinder with the test case given by
    `testName`.

    :param testName: Name of the test case
    :param estimator: The type of estimator to use
    """
    if testName in KNOWN_FAILURES:
        pytest.skip()

    # initialize the algorithm under test
    simpleArgumentAlgo = SimpleArgumentNewtonAlgorithm(estimatorType=estimator)

    context = buildContextFromData(testFunctions[testName])
    simpleArgumentAlgo.calcRoots(context)
    foundRoots = context.container.getRoots()  # pylint: disable=E1111
    expectedRoots = np.array(testFunctions[testName].expectedRoots)

    assert rootsMatchClosely(
        foundRoots, expectedRoots, precision=testFunctions[testName].precision
    )



"""
Module exposing named constants for log levels.

The constants in this module just act as a facade to their counterparts in the
standard library `logging` module.
"""

import logging
from enum import Enum


class LogLevel(Enum):
    """
    Named constants for log levels. Simple wrapper for the constants defined in
    the standard library logging module.
    """

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL



"""
Class PyZEALParserInterface from the package pyzeal_cli.
This module provides a facade for `argparser.ArgumentParser` class from the
standard library `argparse` module. Any access to (CLI-)parsers in this project
should happen only through this interface.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, Tuple, runtime_checkable

from pyzeal.cli.parse_results import (
    InstallTestingParseResults,
    PluginParseResults,
    SettingsParseResults,
)


@runtime_checkable
class PyZEALParserInterface(Protocol):
    """
    Acts as a facade for the standard library `ArgumentParser`.
    """

    def parseArgs(
        self,
    ) -> Tuple[
        SettingsParseResults, PluginParseResults, InstallTestingParseResults
    ]:
        """
        Read command line arguments, parse the read arguments and return them
        wrapped according to the `pyzeal_cli` data contract for parsed command
        line arguments.

        :return: the wrapped parsing results
        """
        ...



"""
Class SimpleArgumentNewtonAlgorithm from the package pyzeal_algorithms.

This module defines a refined version of the `SimpleArgumentAlgorithm` by
supplementing the argument principle with a Newton algorithm once a starting
point has been identified with sufficient accuracy.

Authors:\n
- Philipp Schuette\n
"""

from typing import Tuple, cast

from scipy.optimize import newton  # type: ignore

from pyzeal.algorithms.constants import FOUR_PI, TWO_PI
from pyzeal.algorithms.simple_holo import SimpleArgumentAlgorithm
from pyzeal.utils.root_context import RootContext


class SimpleArgumentNewtonAlgorithm(SimpleArgumentAlgorithm):
    """
    Class representation of a root finding algorithm combining the phase
    interpretation of the argument principle used in `SimpleArgumentAlgorithm`
    with a number of Newton steps once a sufficient refinement depth has been
    reached.
    """

    def decideRefinement(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        phi: float,
        context: RootContext,
    ) -> None:
        """
        Decide which way the current search area should be subdivided and
        calculate roots in the subdivided areas. The simple strategy consists
        of the choices (1) return, if argument indicates no roots, (2) place
        root in `context.container` if roots present and accuracy attained,
        (3) start Newton algorithm if simple root present and sufficent
        accuracy attained, or (4) subdivide further if multiple roots present
        or attained accuracy insufficient for Newton algorithm.

        :param reRan: Real part of current search Range
        :param imRan: Imaginary part of current search range
        :param phi: Change in argument along the boundary of the current range
        :param context: `RootContext` in which the algorithm operates
        """
        # calculate difference between right/left and top/bottom
        x1, x2 = reRan
        y1, y2 = imRan
        deltaRe = x2 - x1
        deltaIm = y2 - y1

        # check if desired accuracy is aquired
        epsReal = 10 ** (-context.precision[0])
        epsImag = 10 ** (-context.precision[1])
        if phi > TWO_PI and deltaRe < epsReal and deltaIm < epsImag:
            SimpleArgumentAlgorithm.getRootFromRectangle(
                x2, x1, y2, y1, phi, context
            )
            return

        # check if the current box contains a simple root - Newton's algorithm
        # does not perform well enough to start within large rectangles
        # TODO: keep on refining if Newton's algorithm finds no (valid) root!
        if TWO_PI < phi < FOUR_PI and deltaRe < 0.1 and deltaIm < 0.1:
            try:
                self.logger.debug(
                    "starting Newton algorithm from %f+%fj", x1, (y1 + y2) / 2
                )
                newRoot = newton(
                    func=context.f,
                    x0=x1 + 0.5 * (y1 + y2) * 1j,
                    x1=x2 + 0.5 * (y1 + y2) * 1j,
                    fprime=None,
                    maxiter=50,
                    tol=min(epsReal, epsImag),
                )
                context.container.addRoot(
                    (cast(complex, newRoot), 1), context.toFilterContext()
                )
                if context.df is not None:
                    newRoot = newton(
                        func=context.f,
                        x0=(x1 + x2 + 1j * (y1 + y2)) * 0.5,
                        fprime=context.df,
                        maxiter=50,
                        tol=min(epsReal, epsImag),
                    )
                    context.container.addRoot(
                        (cast(complex, newRoot), 1), context.toFilterContext()
                    )
            except RuntimeError:
                pass
            finally:
                if context.progress is not None and context.task is not None:
                    context.progress.update(
                        context.task, advance=deltaRe * deltaIm
                    )
            return

        super().decideRefinement((x1, x2), (y1, y2), phi, context)



"""
Class RootFinder from the package pyzeal.

This module defines an implementation of the main root finding API as defined
by the `RootFinderInterface` protocol. The class implemented here provides only
the most basic context for a root finding strategy, e.g. no support for
multiprocessing.

Authors:\n
- Philipp Schuette\n
"""

from typing import Optional, Tuple

import numpy as np
from numpy.typing import NDArray
from rich.progress import TaskID

from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.loggable import Loggable
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.pyzeal_types.root_types import tHoloFunc, tVec
from pyzeal.pyzeal_types.settings_types import SettingsServicesTypes
from pyzeal.rootfinders.finder_interface import RootFinderInterface
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.finder_progress import FinderProgressBar
from pyzeal.utils.root_context import RootContext
from pyzeal.utils.service_locator import ServiceLocator


class RootFinder(RootFinderInterface, Loggable):
    """
    Simple (i.e. non-parallel) implementation of the main root finding API.
    """

    __slots__ = (
        "f",
        "df",
        "algorithm",
        "_container",
        "precision",
        "numSamplePoints",
        "verbose",
    )

    def __init__(
        self,
        f: tHoloFunc,
        df: Optional[tHoloFunc] = None,
        *,
        containerType: ContainerTypes = ContainerTypes.DEFAULT,
        algorithmType: AlgorithmTypes = AlgorithmTypes.DEFAULT,
        estimatorType: EstimatorTypes = EstimatorTypes.DEFAULT,
        precision: Optional[Tuple[int, int]] = None,
        numSamplePoints: Optional[int] = None,
        verbose: Optional[bool] = None,
    ) -> None:
        """
        Initialize a simple, non-parallel root finder.

        :param f: the function whose roots should be calculated
        :param df: the derivative of `f`
        :param containerType: the type of container found roots are stored in
        :param algorithmType: the type of algorithm used for root finding
        :param estimatorType: the type of argument estimator used
        :param precision: the accuracy at which roots are considered exact
        :param numSamplePoints: determines grid size for `NewtonGridAlgorithm`
        :param verbose: flag that toggles the command line progress bar
        """
        self.f = f
        self.df = df
        self.algorithm: FinderAlgorithm = ServiceLocator.tryResolve(
            FinderAlgorithm,
            algoType=algorithmType,
            estimatorType=estimatorType,
            numSamplePoints=numSamplePoints,
        )
        self._container = ServiceLocator.tryResolve(
            RootContainer, containerType=containerType, precision=precision
        )
        self.precision = (
            precision or ServiceLocator.tryResolve(SettingsService).precision
        )

        self.verbose = (
            verbose
            if verbose is not None
            else ServiceLocator.tryResolve(
                SettingsService, settingsType=SettingsServicesTypes.DEFAULT
            ).verbose
        )
        self.logger.debug("initialized the new root finder %s!", str(self))

    def __str__(self) -> str:
        """
        Return a simple string representation of a `RootFinder` instance.
        """
        if self.df is not None:
            return (
                f"RootFinder(f={getattr(self.f, '__name__', '<unnamed>')}, "
                + f"df={getattr(self.df, '__name__', '<unnamed>')})"
            )
        return (
            f"RootFinder(f={getattr(self.f, '__name__', '<unnamed>')}, "
            + "df=None)"
        )

    def calculateRoots(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        precision: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Start a (non-parallel) root finding calculation in the rectangle
        `reRan x imRan` up to a number of `precision` significant digits in
        real and imaginary part.

        :param reRan: horizontal extend of the complex region to search in
        :param imRan: vertical extend of the complex region to search in
        :param precision: accuracy of the search in real and imaginary parts
        """
        # if no precision was given, use default precision from constructor
        precision = precision or self.precision
        # if a rounding container is used we must calculate with an additional
        # digit of internal precision to obtain correct results after rounding
        precision = (precision[0] + 1, precision[1] + 1)
        # desymmetrize the input rectangle
        (x1, x2), (y1, y2) = self.desymmetrizeDomain(reRan, imRan, precision)

        # initialize the progress bar
        progress = FinderProgressBar() if self.verbose else None
        task: Optional[TaskID] = None
        if progress is not None:
            task = progress.addTask((x2 - x1) * (y2 - y1))
            progress.start()
            self.logger.debug("starting progress bar...")

        # construct the root finding context
        context = RootContext(
            f=self.f,
            df=self.df,
            container=self.container,
            precision=precision,
            reRan=(x1, x2),
            imRan=(y1, y2),
            progress=progress,
            task=task,
        )
        # shut down root finding in orderly fashion upon command line signals
        try:
            self.logger.info("attempting to calculate roots...")
            self.algorithm.calcRoots(context)
            if progress is not None and task is not None:
                progress.update(task, description="[green] search finished!")
        except KeyboardInterrupt:
            self.logger.warning(
                "root calculation interrupted - some roots may be missing!"
            )
            if progress and task:
                progress.stop_task(task)
                progress.update(task, visible=False)
                progress.refresh()
        if progress is not None and task is not None:
            progress.stop()
        self.logger.info("non-parallel root search finished!")

    @property
    def roots(self) -> tVec:
        """
        Return the roots calculated with this root finder through previous
        calls to `calculateRoots()`.

        :return: the set of roots calculated by this finder
        """
        return self.container.getRoots()

    @property
    def orders(self) -> NDArray[np.int32]:
        """
        Return the orders of the roots calculated with this finder. The output
        is parallel to the return value of the `roots` property.

        :return: the orders of the roots calculated by this finder
        """
        return self.container.getRootOrders()

    def setLevel(self, level: LogLevel) -> None:
        """
        Set the logging level of this `RootFinder` and all dependent objects
        like containers and algorithms.

        :param level: the new logging level
        """
        super().setLevel(level)
        self.container.setLevel(level)
        self.algorithm.setLevel(level)

    @property
    def container(self) -> RootContainer:
        """
        Return the container attached to this rootfinder
        """
        return self._container

    def desymmetrizeDomain(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        precision: Tuple[int, int],
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """
        Desymmetrize the domain to improve numerical stability.

        In practice one notices that target domains as given by `reRan x imRan`
        often contain roots on/near their boundaries if given e.g. in terms of
        integers. This method is automatically called by `calculateRoots` so a
        user does not need to worry about (slightly) perturbing their domains.

        :param reRan: Search range for the real part
        :param imRan: Search range for the imaginary part
        :param precision: Accuracy in real and imaginary part
        :return: New bounds for real and imaginary parts
        """
        (x1, x2), (y1, y2) = sorted(reRan), sorted(imRan)
        x1 = x1 - 1 * 10 ** (-1 * precision[0])
        x2 = x2 + 2 * 10 ** (-1 * precision[0])
        y1 = y1 - 3 * 10 ** (-1 * precision[1])
        y2 = y2 + 4 * 10 ** (-1 * precision[1])
        self.logger.debug(
            "desymmetrized root finding domain to [%f, %f] x [%f, %f]",
            x1,
            x2,
            y1,
            y2,
        )
        return (x1, x2), (y1, y2)



"""
TODO.

Authors:\n
- Philipp Schuette\n
"""

from json import JSONDecodeError, dump, load
from typing import Dict, Literal, Set, Tuple, Type, Union

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.invalid_setting_exception import InvalidSettingException

# admissible keys when changing settings
tCoreSettingsKey = Union[
    Literal["defaultContainer"],
    Literal["defaultAlgorithm"],
    Literal["defaultEstimator"],
]
tSettingsKey = Union[
    Literal["logLevel"],
    Literal["verbose"],
    Literal["precision"],
]
# admissible types when assigning to settings properties
tCoreSettingsPropertyType = Union[
    ContainerTypes, AlgorithmTypes, EstimatorTypes
]
tSettingsPropertyType = Union[LogLevel, bool, Tuple[int, int]]


class JSONHelper:
    """
    Static helper class providing methods to read and manipulate json settings.
    """

    _coreSettings: Dict[
        tCoreSettingsKey,
        Union[
            Type[AlgorithmTypes], Type[ContainerTypes], Type[EstimatorTypes]
        ],
    ] = {
        "defaultAlgorithm": AlgorithmTypes,
        "defaultContainer": ContainerTypes,
        "defaultEstimator": EstimatorTypes,
    }
    _settings: Set[tSettingsKey] = {"logLevel", "verbose", "precision"}

    @staticmethod
    def createOrUpdateCoreSetting(
        filename: str,
        setting: tCoreSettingsKey,
        value: tCoreSettingsPropertyType,
    ) -> None:
        """
        Update a core setting or create a new one if no value had previously
        been set.

        :param filename: the json file to update
        :param setting: setting to create or update
        :param value: New setting value
        :raises InvalidSettingException: If the given value is invalid for the
            specified setting, an `InvalidSettingException` is raised.
        """
        currentSettings: Dict[tCoreSettingsKey, str] = {}
        try:
            with open(filename, "r", encoding="utf-8") as custom:
                currentSettings = load(custom)
        except FileNotFoundError:
            pass
        except JSONDecodeError as exc:
            raise InvalidSettingException(f"{filename=}") from exc

        if setting in JSONHelper._coreSettings:
            if not isinstance(value, JSONHelper._coreSettings[setting]):
                raise InvalidSettingException(setting)
            currentSettings[setting] = value.value
        else:
            raise InvalidSettingException(f"key={setting}")

        with open(filename, "w", encoding="utf-8") as custom:
            dump(currentSettings, custom, indent=4)

    @staticmethod
    def loadCoreSettingsFromFile(
        filename: str, settings: Dict[str, str]
    ) -> None:
        """
        Load the settings stored in `filename` into `settings`.

        :param filename: Settings file to load
        :param settings: Dict to store the read settings in
        """
        try:
            with open(filename, "r", encoding="utf-8") as settingsFile:
                for key, value in load(settingsFile).items():
                    if key not in JSONHelper._coreSettings:
                        continue
                    settings[key] = value
        except FileNotFoundError:
            pass

    @staticmethod
    def createOrUpdateSetting(
        filename: str, setting: tSettingsKey, value: tSettingsPropertyType
    ) -> None:
        """
        Update a setting or create a new setting if no value has been set yet.

        :param filename: the json file to update
        :param setting: setting to create or update
        :param value: New setting value
        :raises InvalidSettingException: If the given value is invalid for the
            specified setting, an `InvalidSettingException` is raised.
        """
        currentSettings: Dict[
            tSettingsKey, Union[str, bool, Tuple[int, int]]
        ] = {}
        try:
            with open(filename, "r", encoding="utf-8") as custom:
                currentSettings = load(custom)
        except FileNotFoundError:
            pass
        except JSONDecodeError as exc:
            raise InvalidSettingException(f"{filename=}") from exc

        if setting == "logLevel":
            if not isinstance(value, LogLevel):
                raise InvalidSettingException("default algorithm")
            currentSettings["logLevel"] = value.name
        elif setting == "verbose":
            if not isinstance(value, bool):
                raise InvalidSettingException("default verbosity")
            currentSettings["verbose"] = value
        elif setting == "precision":
            if not (
                isinstance(value, tuple)
                and isinstance(value[0], int)
                and isinstance(value[1], int)
            ):
                raise InvalidSettingException("default precision")
            currentSettings["precision"] = (value[0], value[1])
        else:
            raise InvalidSettingException(f"key={setting}")

        with open(filename, "w", encoding="utf-8") as custom:
            dump(currentSettings, custom, indent=4)

    @staticmethod
    def loadSettingsFromFile(
        filename: str, settings: Dict[str, Union[str, bool, Tuple[int, int]]]
    ) -> None:
        """
        Load the settings stored in `filename` into `settings`.

        :param filename: File to load
        :param settings: Dict to store the read settings in
        """
        try:
            with open(filename, "r", encoding="utf-8") as settingsFile:
                for key, value in load(settingsFile).items():
                    if key not in JSONHelper._settings:
                        continue
                    if isinstance(value, list):
                        value = tuple(value)
                    settings[key] = value
        except FileNotFoundError:
            pass
